package org.openxma.dsl.common.formatter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.AbstractRule;
import org.eclipse.xtext.GrammarUtil;
import org.eclipse.xtext.ParserRule;
import org.eclipse.xtext.RuleCall;
import org.eclipse.xtext.conversion.IValueConverterService;
import org.eclipse.xtext.formatting.IFormatter;
import org.eclipse.xtext.formatting.impl.AbstractNodeModelFormatter;
import org.eclipse.xtext.parsetree.AbstractNode;
import org.eclipse.xtext.parsetree.CompositeNode;
import org.eclipse.xtext.parsetree.LeafNode;
import org.eclipse.xtext.parsetree.SyntaxError;
import org.eclipse.xtext.parsetree.reconstr.impl.TokenStringBuffer;
import org.openxma.dsl.common.formatter.OpenXMAFormatter.CustomOneWhitespaceFormatterStream;

import com.google.inject.Inject;

public class OpenXMANodeModelFormatter extends AbstractNodeModelFormatter {
	
	@Inject
	protected IFormatter formatter;

	@Inject
	protected IValueConverterService valueConverter;

	@Override
	public IFormattedRegion format(CompositeNode root, int offset, int length) {
		List<AbstractNode> nodes = getRelevantNodes(root, offset, offset + length);
		if (nodes.size() == 0)
			return null;

		String indent = getIndentation(root, offset);
		TokenStringBuffer buf = new TokenStringBuffer();
		CustomOneWhitespaceFormatterStream fmt = (CustomOneWhitespaceFormatterStream)formatter.createFormatterStream(indent, buf, false);
		try {
			for (AbstractNode currentNode : nodes) {
				EObject grammarElement = currentNode.getGrammarElement();							
				// LeafNode
				if (currentNode instanceof LeafNode) {
					LeafNode l = (LeafNode) currentNode;
					if (l.isHidden()) {
						fmt.writeHidden(grammarElement, l.getText());
					} else {
						fmt.writeSemantic(grammarElement, l.getText());
					}
					continue;
				}								
				// DatatypeRuleCall
				if (isDatatypeRuleCall(currentNode)) {				
					AbstractRule rule = getRuleOfRuleCall(grammarElement);						
					String ruleName = rule.getName();					
					Object val = valueConverter.toValue(nodeToStr((CompositeNode) currentNode),ruleName, currentNode);
					String text = valueConverter.toString(val,ruleName);
					fmt.initGrammarElement(currentNode,grammarElement);						
					fmt.writeSemantic(rule,text);							
					continue;
				}				
				// Other RuleCall then DataTypeRuleCall
				if (grammarElement instanceof RuleCall) {
					RuleCall rc = (RuleCall)grammarElement;
					AbstractRule rule = (AbstractRule)rc.getRule();
					fmt.initGrammarElement(currentNode,grammarElement);
					fmt.initGrammarElement(currentNode,rule);					
					continue;
				}				
				// Other
				fmt.initGrammarElement(currentNode,grammarElement);											
			}
			fmt.flush();
		} catch (IOException e) {
			// this should never happen since TokenStringBuffer doesn't throw
			// IOEs.
			throw new RuntimeException(e);
		}
		int start = nodes.get(0).getOffset();
		int lengt = (nodes.get(nodes.size() - 1).getOffset() + nodes.get(
				nodes.size() - 1).getLength())
				- start;
		return new FormattedRegion(start, lengt, buf.toString().trim());
	}

	protected String getIndentation(EObject root, int fromOffset) {
		if (fromOffset == 0)
			return "";

		List<LeafNode> r = new ArrayList<LeafNode>();
		Iterator<EObject> it = root.eAllContents();

		// add all nodes until fromOffset
		while (it.hasNext()) {
			EObject o = it.next();
			if (o instanceof LeafNode) {
				LeafNode l = (LeafNode) o;
				if (l.getOffset() >= fromOffset)
					break;
				else
					r.add(l);
			}
		}

		// go backwards until first linewrap
		Pattern p = Pattern.compile("\\n([ \\t]*)");
		for (int i = r.size() - 1; i >= 0; i--) {
			Matcher m = p.matcher(r.get(i).getText());
			if (m.find()) {
				String ind = m.group(1);
				while (m.find())
					ind = m.group(1);
				return ind;
			}
		}
		return "";
	}
	
	protected List<AbstractNode> getRelevantNodes(EObject root, int fromOffset,
			int toOffset) {
		List<AbstractNode> r = new ArrayList<AbstractNode>();
		TreeIterator<EObject> it = root.eAllContents();

		// seek to fromOffset
		if (fromOffset > 0)
			while (it.hasNext()) {
				EObject o = it.next();
				boolean dt = isDatatypeRuleCall(o);
				if (dt) {
					it.prune();
				}
				if (o instanceof SyntaxError ){
					throw new RuntimeException(o.toString());
				}				
				AbstractNode l = (AbstractNode) o;
				if (l.getOffset() + l.getLength() >= fromOffset) {
					r.add(l);
					break;
				}
			}

		// add Leafs within the range
		while (it.hasNext()) {
			EObject o = it.next();
			boolean dt = isDatatypeRuleCall(o);
			if (dt) {
				it.prune();
			}
			if (o instanceof SyntaxError ){
				throw new RuntimeException(o.toString());
			}
			AbstractNode l = (AbstractNode) o;
			if (l.getOffset() > toOffset)
				break;
			else
				r.add(l);						
		}

		// remove tailing hidden leafs
		for (int i = r.size() - 1; i >= 0; i--)
			if (r.get(i) instanceof LeafNode
					&& ((LeafNode) r.get(i)).isHidden())
				r.remove(i);
			else
				break;
		return r;
	}
	

	protected List<AbstractNode> getLeafs(EObject root, int fromOffset,
			int toOffset) {
		List<AbstractNode> r = new ArrayList<AbstractNode>();
		TreeIterator<EObject> it = root.eAllContents();

		// seek to fromOffset
		if (fromOffset > 0)
			while (it.hasNext()) {
				EObject o = it.next();
				boolean dt = isDatatypeRuleCall(o);
				boolean ln = o instanceof LeafNode
						&& !((LeafNode) o).isHidden();
				if (dt)
					it.prune();
				if (dt || ln) {
					AbstractNode l = (AbstractNode) o;
					if (l.getOffset() + l.getLength() >= fromOffset) {
						r.add(l);
						break;
					}
				}
			}

		// add Leafs within the range
		while (it.hasNext()) {
			EObject o = it.next();
			boolean dt = isDatatypeRuleCall(o);
			if (dt)
				it.prune();
			if (o instanceof LeafNode || dt) {
				AbstractNode l = (AbstractNode) o;
				if (l.getOffset() > toOffset)
					break;
				else
					r.add(l);
			}
		}

		// remove tailing hidden leafs
		for (int i = r.size() - 1; i >= 0; i--)
			if (r.get(i) instanceof LeafNode
					&& ((LeafNode) r.get(i)).isHidden())
				r.remove(i);
			else
				break;
		return r;
	}
	
	

	protected boolean isDatatypeRuleCall(EObject node) {
		if (node instanceof CompositeNode
				&& ((CompositeNode) node).getGrammarElement() instanceof RuleCall) {
			RuleCall rc = (RuleCall) ((CompositeNode) node).getGrammarElement();
			if (rc.getRule() instanceof ParserRule)
				return GrammarUtil.isDatatypeRule((ParserRule) rc.getRule());
		}
		return false;
	}
	
	protected boolean isLeafNode(EObject o) {
		return o instanceof LeafNode;
	}

	protected String nodeToStr(CompositeNode node) {
		Iterator<EObject> it = node.eAllContents();
		StringBuffer text = new StringBuffer();
		StringBuffer hidden = new StringBuffer();
		while (it.hasNext()) {
			EObject o = it.next();
			if (o instanceof LeafNode) {
				LeafNode l = (LeafNode) o;
				if (l.isHidden())
					hidden.append(l.getText());
				else {
					if (text.length() > 0)
						text.append(hidden);
					hidden.delete(0, hidden.length());
					text.append(l.getText());
				}
			}
		}
		return text.toString();
	}
	
	public static AbstractRule getRuleOfRuleCall(EObject grammarElement) {
		if (grammarElement instanceof RuleCall) {
			RuleCall ruleCall = (RuleCall)grammarElement;
			return ruleCall.getRule();							
		} 
		return null;
	}
	

}
